
/* Copyright (c) 1994-95 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 * Author: Satish Chandra, 1994
 * NOTE: This code was derived from a non copyrighted version from
 * Syracuse University. Original author unknown.
 */

#include "f2c.h"
#include <cm/cmmd/io.h>
#include <cm/cmmd.h>

/* Common Block Declarations */

integer n, nn, numcol, numcollg, mincol, maxcol;
integer np, bb, numrow;

integer numnode, numdim, thisnode, hostnode, env[4];

struct {
    integer seed;
} random_seed__;

#define random_seed__1 random_seed__

/* Initialized data */

#define gtype  999
#define realsize 4

#define MAX_ROW 512
#define MAX_COL 512

#define MAX_AUG_COL (MAX_COL + 1)

#define MAX_SIZE (MAX_ROW * MAX_AUG_COL)

real a[MAX_SIZE];
real aa[MAX_SIZE];
integer index[MAX_ROW];
real sum[MAX_ROW];
real fac[MAX_ROW];
integer mask[MAX_ROW], mask2[MAX_ROW];
real b[MAX_ROW];
real sendbuf[MAX_AUG_COL]; 

/* Main program */ MAIN__(void)
{
    extern integer cmmd_self_address_(void);
    extern /* Subroutine */ int init_1_bcast_(integer *, integer *);
    extern integer cmmd_partition_size_(void);

    extern /* Subroutine */ int getdata_(real *, integer *), backsub_(real *, 
	    integer *, integer *), gaussit_(real *, real *, integer *, 
	    integer *);

    thisnode = cmmd_self_address_();
    n = 512;
/*       expect N to be 32 * k */
    np = n + 1;
    numnode = cmmd_partition_size_();
    numrow = n / numnode;
    init_1_bcast_(&thisnode, &numnode);
    getdata_(a, &numrow);
/*       Print VT1 */
#ifdef WWT
      {
         int time;
         extern int wwt_node_number;

         time = wwt_get_vt_lo();
         
         if (wwt_node_number == 0) printf("VT now = %d\n", time);
      }
#else
#endif
    gaussit_(a, aa, &numrow, index);
/*       Print VT3 */
#ifdef WWT
      {
         int time;
         extern int wwt_node_number;

         time = wwt_get_vt_lo();
         
         if (wwt_node_number == 0) printf("VT now = %d\n", time);
      }
#else
#endif
    backsub_(aa, index, &numrow);
    return 0;
} /* MAIN__ */

/* Subroutine */ int getdata_(real *y, integer *m)
{
    /* System generated locals */
    integer y_dim1, i__1, i__2;

    /* Local variables */
    integer i, j, k;
    extern /* Subroutine */ int initrandom_(integer *);

    extern doublereal srandom_(void);
    integer npl = np;

#define y_ref(a_1,a_2) y[(a_1)*y_dim1 + a_2]

/*       the following areas are user defined. */
/*       this area is system area. */
/* 	local variables. */
    /* Parameter adjustments */
    y_dim1 = npl;

    /* Function Body */
    random_seed__1.seed = (thisnode + 1) * 100000000;
    initrandom_(&random_seed__1.seed);
    i__1 = numrow - 1;
    for (i = 0; i <= i__1; ++i) {
	i__2 = npl - 1;
	for (j = 0; j <= i__2; ++j) {
	    y_ref(i, j) = srandom_() * srandom_() / 2.f;
	}
    }
    i__1 = numrow - 1;
    for (i = 0; i <= i__1; ++i) {
	sum[i] = y_ref(i, 0);
	k = 2;
	i__2 = n - 1;
	for (j = 1; j <= i__2; ++j) {
	    sum[i] += k * y_ref(i, j);
	    ++k;
	}
	y_ref(i, npl - 1) = sum[i];
    }
    return 0;
} /* getdata_ */

#undef y_ref

/* Subroutine */ int gaussit_(real *y, real *z, integer *m, integer *index)
{
    /* System generated locals */
    integer y_dim1, z_dim1, i__1, i__2, i__3;
    real r__1;
    integer ntype = 999;

    /* Local variables */
    real localmax;
    integer i, k, j, ro;
    extern /* Subroutine */ int do_1_gshigh_(real *, real *, integer *);
    extern /* Subroutine */ int do_broadcast_(integer *, real *, integer *, 
	    integer *, integer *);
    integer start;
    extern /* Subroutine */ int get_broadcast_(integer *, real *, integer *, 
	    integer *, integer *);

    extern float SC_reduce_float(float, int);
    real maxval;
    integer maxrow;
    real tempmax;

    integer counter;
    integer col;
    integer npl = np;
    real *sendrow;


#define y_ref(a_1,a_2) y[(a_1)*y_dim1 + a_2]
#define z_ref(a_1,a_2) z[(a_1)*z_dim1 + a_2]

    /* Parameter adjustments */
    z_dim1 = npl;
    y_dim1 = npl;

    /* Function Body */
    counter = 0;
    i__1 = *m - 1;
    for (i = 0; i <= i__1; ++i) {
	mask[i] = -1;
	mask2[i] = -1;
    }
/*       STEP THROUGH ALL THE COLUMNS. */
    i__1 = n - 2;
    for (col = 0; col <= i__1; ++col) {
/* 	PHASE 1 COMPUTATION. */
	maxval = -1.f;
	localmax = -1.f;
	maxrow = 0;
	i__2 = *m - 1;
	for (ro = 0; ro <= i__2; ++ro) {
	    if (mask[ro] == -1 && (maxval < dabs(y_ref(ro, col)))) 
	      {
		maxval = dabs(y_ref(ro, col));
		maxrow = ro;
	      }
	}
	if (maxval != -1.f) {
	    maxval = maxval * 1e6f + thisnode;
	}
	localmax = maxval;
/*       FIND THE MAXIMUM OF ALL PIVOTS FOUND BY EACH NODE */
#ifdef AM_REDUC
	maxval = SC_reduce_float(maxval, CMMD_combiner_fmax);
#else
#ifndef SBC
	do_1_gshigh_(&maxval, &tempmax, &thisnode);
#else
	i__3 = 1;
	do_gshigh_(&ntype, &maxval, &i__3, &tempmax, &thisnode, &numnode);
#endif
#endif /* AM_REDUC */
	if (maxval != -1.f) {
	    if (maxval == localmax) {
/*       Print High3 */
		maxval = y_ref(maxrow, col);
		k = maxrow;
/*       sendRow CONTAINS THE PIVOT ROW. */
		/*
		i__2 = npl - 1;
		for (j = col; j <= i__2; ++j) {
		    sendrow[j - col] = y_ref(k, j);
		}
		*/
		sendrow = &(y_ref(k, col));

/*       sendRow IS SENT TO REMAINING NODES. */
		i__2 = gtype + (col << 1);
		i__3 = realsize * (npl - col);
#ifdef SBC
		do_broadcast_(&i__2, sendrow, &i__3, &thisnode, &
			      numnode);
#else
		CMMD_bc_to_nodes(sendrow, i__3);
#endif
		mask[maxrow] = col;
		mask2[maxrow] = counter;
		++counter;
	    } else {
/*       sendRow IS RECEIVED BY THE REST OF THE NODES. */
		i__2 = gtype + (col << 1);
		i__3 = realsize * (npl - col);
		sendrow = &(sendbuf[0]);
#ifndef SBC		
		CMMD_receive_bc_from_node(sendrow, i__3);
#else
		get_broadcast_(&i__2, sendrow, &i__3, &thisnode, &
			       numnode);
#endif

		
	    }
	} else {
	    if (maxval == -1.f) {
		maxval = y_ref(maxrow, col);
		k = maxrow;
		i__2 = npl - 1;
		for (j = col; j <= i__2; ++j) {
		    sendrow[j - col] = y_ref(k, j);
		}
		mask[maxrow] = col;
		mask2[maxrow] = counter;
		++counter;
	    }
	}
/*       Row updating */
/* 	PHASE 3 COMPUTATION */
	start = col;
	i__2 = *m - 1;
	for (i = 0; i <= i__2; ++i) {
	    fac[i] = y_ref(i, start) / sendrow[0];
	}
	i__2 = *m -1;
	i__3 = npl - 1;
	for (i = 0;i <= i__2;++i)
	  if (mask[i] == -1)
	    for (j = start; j <= i__3; ++j)
	      y_ref(i, j) = y_ref(i, j) - sendrow[j - col] *  fac[i];
      }
/*       TRIANGULARIZATION. */
/*       PHASE 4 COMPUTATION. */
    i__1 = *m - 1;
    for (i = 0; i <= i__1; ++i) {
	if (mask2[i] == -1) {
	    mask2[i] = counter;
	}
	if (mask[i] == -1) {
	    mask[i] = col;
	}
	k = mask2[i];
	index[k] = mask[i];
	i__2 = npl - 1;
	for (j = 0; j <= i__2; ++j) {
	    z_ref(k, j) = y_ref(i, j);
	}
    }
    return 0;
} /* gaussit_ */

#undef z_ref
#undef y_ref


/* Subroutine */ int backsub_(real *a, integer *mask, integer *m)
{
    /* System generated locals */
    integer a_dim1, i__1, i__2, i__3;

    /* Local variables */

    integer i;
    real x;
    extern /* Subroutine */ int do_1_bcast_(real *, integer *), get_1_bcast_(
	    real *, integer *);
    extern /* Subroutine */ void SC_bc_to_nodes_float(float);
    extern /* Subroutine */ void SC_receive_bc_float(float *);
    integer row, col;


#define a_ref(a_1,a_2) a[(a_1)*a_dim1 + a_2]

    /* Parameter adjustments */
    a_dim1 = np;

    /* Function Body */
    i__1 = *m - 1;
    for (i = 0; i <= i__1; ++i) {
	b[i] = a_ref(i, n);
    }
    row = *m - 1;
    for (col = n - 1; col >= 0; --col) {
	if (row >= 0 && mask[row] == col) {
	    b[row] /= a_ref(row, col);
	    x = b[row];
	    /* do_1_bcast_(&x, &thisnode); */
#ifdef AM_BROAD
	    SC_bc_to_nodes_float(x);
#else
#ifdef SBC
	    i__2 = 0; i__3 = 4;
	    do_broadcast_(&i__2, &x, &i__3, &thisnode, &
			   numnode);
	    
#else
	    CMMD_bc_to_nodes(&x,4);
#endif
#endif /* AM_BROAD */
	    --row;
	} else {
	  /* all other nodes receive the most recently computed x. */
	  /* get_1_bcast_(&x, &thisnode); */
#ifdef AM_BROAD
	  SC_receive_bc_float(&x);
#else
#ifdef SBC
	  i__2 = 0; i__3 = 4;
	  get_broadcast_(&i__2, &x, &i__3, &thisnode, &
			 numnode);
#else		
	  CMMD_receive_bc_from_node(&x,4);
#endif
#endif /* AM_BROAD */
	}
	if (row >= 0) {
	    for (i = row; i >= 0; --i) {
		b[i] -= a_ref(i, col) * x;
	    }
	}
    }
/*       Print VT2 */
#ifdef WWT
      {
         int time;
         extern int wwt_node_number;

         time = wwt_get_vt_lo();
         
         if (wwt_node_number == 0) printf("VT now = %d\n", time);
      }
#else
#endif
/* 	end if */
    i__1 = numrow - 1;
    for (i = 0; i <= i__1; ++i) {
/* 	  Print Result */
            printf("b[%d] = %f\n", mask[i], b[i]);
    }
    return 0;
} /* backsub_ */

#undef a_ref


/* Main program alias */ int gaussnode_ () { MAIN__ (); return 0; }
/*
#ifndef WWT
pn_main()
#else
main()
#endif
{
  gaussnode_();
}
*/
main()
{
#ifndef WWT
 CMMD_fset_io_mode(stdout, CMMD_independent);
#endif
 CMMD_node_timer_start(0);
 gaussnode_();
 CMMD_node_timer_stop(0);
 if (thisnode == 0) printf("Elapsed time = %f sec\n", CMMD_node_timer_elapsed(0));
 }
